﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Data
Imports System.Data.SqlClient
Imports System.Data.SqlTypes
Imports Microsoft.SqlServer.Server

Imports System.Collections.Generic

' this project creates and deploys a CLR stored procedeure that does 
' graph shortest path analysis


Partial Public Class StoredProcedures
    <Microsoft.SqlServer.Server.SqlProcedure()>
    Public Shared Sub csp_ShortestPath(ByVal startNode As System.Data.SqlTypes.SqlInt64, ByVal endNode As SqlInt64, ByVal maxNodesToCheck As SqlInt32, <System.Runtime.InteropServices.Out()> ByRef pathResult As SqlString, <System.Runtime.InteropServices.Out()> ByRef distResult As SqlDouble)
        ' Put your code here. OK, I will!
        Dim distance As New Dictionary(Of Long, Double)() ' curr shortest distance (double) from startNode to (key) node
        Dim previous As New Dictionary(Of Long, Long)() ' predecessor values to construct the actual path when done
        Dim beenAdded As New Dictionary(Of Long, Boolean)() ' has node been added to the queue yet? Once added, we don't want to add again
        Dim PQ As New MyPriorityQueue()

        Dim conn As SqlConnection = Nothing

        Dim startNodeAsLong As Long = CLng(Fix(startNode))
        Dim endNodeAsLong As Long = CLng(Fix(endNode))
        Dim maxNodesToCheckAsInt As Integer = CInt(Fix(maxNodesToCheck))

        ' initialize start node
        distance(startNodeAsLong) = 0.0 ' distance from start node to itself is 0
        previous(startNodeAsLong) = -1 ' -1 is the code for undefined.
        PQ.Enqueue(New NodeDistance(startNodeAsLong, 0.0)) ' the queue holds distance information because Enqueue and Dequeue need it to keep queue ordered by priority. alt approach would be to store only node IDs and then do a distance lookup
        beenAdded(startNodeAsLong) = True

        Try
            'string connString = "context connection=true";  // the sp is running on the db server so we use a local connection but MultipleActiveResultSets=True not allowed
            Dim connString As String = "Server=(local);Database=dbShortPathWithCLR;Trusted_Connection=True;MultipleActiveResultSets=True" ' to allow two datareaders

            conn = New SqlConnection(connString) ' Note: the 'using()' construction seems to have a bad perf impact so we explicitly open and close
            conn.Open()

            Dim alt As Double = 0.0 ' 'test distance'

            Do While PQ.Count() > 0 AndAlso beenAdded.Count < maxNodesToCheckAsInt
                Dim u As NodeDistance = PQ.Dequeue() ' node with shortest distance from start
                If u.nodeID = endNode Then ' found the target end node
                    Exit Do
                End If

                ' fetch all neighbors (v) of u
                Dim cmd As New SqlCommand("select toNode from tblGraph where fromNode=" & u.nodeID)
                cmd.Connection = conn ' cmd.CommandTimeout can be set here

                Dim v As Long ' ID of a neighbor to u
                Dim sdr As SqlDataReader = cmd.ExecuteReader()
                Do While sdr.Read() = True ' if there are no neighbors, this loop will exit immediately
                    v = Long.Parse(sdr(0).ToString())

                    If beenAdded.ContainsKey(v) = False Then ' first appearance of node v
                        distance(v) = Double.MaxValue ' stand-in for infinity
                        previous(v) = -1 ' undefined
                        PQ.Enqueue(New NodeDistance(v, Double.MaxValue)) ' maxValue is a dummy value
                        beenAdded(v) = True
                    End If

                    Dim distCmd As New SqlCommand("select edgeWeight from tblGraph where fromNode=" & u.nodeID & " and toNode=" & v)
                    distCmd.Connection = conn
                    Dim d As Double = Convert.ToDouble(distCmd.ExecuteScalar())

                    'alt = distance[u.nodeID] + 1.0;  // if all edge weights are just hops can simplify to this
                    alt = distance(u.nodeID) + d ' alt = dist(start-to-u) + dist(u-to-v), so alt is total distance from start to v

                    If alt < distance(v) Then ' is alt a new, shorter distance from start to v?
                        distance(v) = alt
                        previous(v) = u.nodeID
                        PQ.ChangePriority(v, alt) ' this will change the distance/priority
                    End If

                Loop ' sdr Read loop
                sdr.Close()

            Loop ' main while loop on PQ.Count > 0

            conn.Close() ' no more selects needed

            ' extract the shortest path as a string from the previous[] array
            pathResult = "NOTREACHABLE" ' default result string
            distResult = -1.0 ' distance result defaults to -1 == unreachable

            If distance.ContainsKey(endNodeAsLong) = True Then ' endNode was encountered at some point in the algorithm
                Dim sp As Double = distance(endNodeAsLong) ' shortest path distance
                If sp <> Double.MaxValue Then ' we have a reachable path
                    pathResult = ""
                    Dim currNode As Long = endNodeAsLong
                    Do While currNode <> startNodeAsLong ' construct the path as semicolon delimited string
                        pathResult += currNode.ToString() & ";"
                        currNode = previous(currNode)
                    Loop
                    pathResult += startNode.ToString() ' tack on the start node
                    distResult = sp ' the distance result
                End If
            End If
            ' try

        Catch ex As Exception ' typically Out Of Memory or a Command TimeOut
            pathResult = ex.Message
            distResult = -1.0

        Finally
            If conn IsNot Nothing AndAlso conn.State = ConnectionState.Open Then
                conn.Close()
            End If
        End Try


    End Sub ' csp_ShortestPath()
End Class ' class StoredProcedures

' =================================

Public Class NodeDistance
    Public nodeID As Long
    Public distance As Double

    Public Sub New(ByVal nodeID As Long, ByVal distance As Double)
        Me.nodeID = nodeID
        Me.distance = distance
    End Sub

    Public Overrides Function ToString() As String
        Return "[ " & nodeID & " " & distance.ToString("F1") & " ]"
    End Function
End Class

Public Class MyPriorityQueue
    Public list As List(Of NodeDistance)

    Public Sub New()
        Me.list = New List(Of NodeDistance)()
    End Sub

    Private Function IndexOf(ByVal nodeID As Long) As Integer
        For i As Integer = 0 To list.Count - 1
            If list(i).nodeID = nodeID Then
                Return i
            End If
        Next i
        Return -1
    End Function

    Private Function IndexOfSmallestDist() As Integer
        Dim smallDist As Double = Me.list(0).distance
        Dim smallIndex As Integer = 0
        For i As Integer = 0 To list.Count - 1
            If list(i).distance < smallDist Then
                smallDist = list(i).distance
                smallIndex = i
            End If
        Next i
        Return smallIndex
    End Function

    Public Function Dequeue() As NodeDistance
        Dim i As Integer = IndexOfSmallestDist()
        Dim result As NodeDistance = list(i)
        list.RemoveAt(i)
        Return result
    End Function

    Public Sub Enqueue(ByVal nd As NodeDistance)
        list.Add(nd)
    End Sub

    Public Sub ChangePriority(ByVal nodeID As Long, ByVal newDist As Double)
        Dim i As Integer = IndexOf(nodeID)
        list(i).distance = newDist
    End Sub

    Public Function Count() As Integer
        Return Me.list.Count
    End Function

    Public Overrides Function ToString() As String
        Dim s As String = ""
        For i As Integer = 0 To list.Count - 1
            s &= list(i).ToString & " "
        Next i
        Return s
    End Function

End Class ' MyPriorityQueue